import asyncio
from abc import ABC, abstractmethod
from typing import Any, Optional

from privex.helpers.exceptions import CacheNotFound

from privex.helpers.cache.CacheAdapter import CacheAdapter
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT
from privex.helpers.types import VAL_FUNC_CORO


class AsyncCacheAdapter(CacheAdapter, ABC):
    """
    **AsyncCacheAdapter** is an abstract base class based on :class:`.CacheAdapter`, but with all methods designated as coroutines.
    
    Cache adapters which make use of AsyncIO, including via asyncio compatible libraries (e.g. ``aioredis``), should use this class
    as their parent instead of :class:`.CacheAdapter`.
    
    To retain the functionality of :meth:`.__getitem__` and :meth:`.__setitem__`, it obtains an event loop
    using :func:`asyncio.get_event_loop`, and then wraps :meth:`.get` or :meth:`.set` respectively using
    ``loop.run_until_complete`` to be able to run them within the synchronous get/setitem magic methods.
    
    It overrides :meth:`.get_or_set` to convert it into an async method, and overrides :meth:`.get_or_set_async` so that
    :meth:`.get` and :meth:`.set` are correctly awaited within the method.
    
    """
    async def get_or_set(self, key: str, value: VAL_FUNC_CORO, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        return await self.get_or_set_async(key=key, value=value, timeout=timeout)

    async def get_or_set_async(self, key: str, value: VAL_FUNC_CORO, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        try:
            k = await self.get(key, fail=True)
        except CacheNotFound:
            k = value
            
            if asyncio.iscoroutinefunction(value):
                k = await value(key)
            if asyncio.iscoroutine(value):
                k = await value
            elif callable(value):
                k = value(key)
            await self.set(key=key, value=k, timeout=timeout)
        return k
    
    @abstractmethod
    async def get(self, key: str, default: Any = None, fail: bool = False) -> Any:
        raise NotImplemented(f'{self.__class__.__name__} must implement .get()')

    @abstractmethod
    async def set(self, key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
        raise NotImplemented(f'{self.__class__.__name__} must implement .set()')

    @abstractmethod
    async def remove(self, *key: str) -> bool:
        raise NotImplemented(f'{self.__class__.__name__} must implement .remove()')

    @abstractmethod
    async def update_timeout(self, key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        raise NotImplemented(f'{self.__class__.__name__} must implement .extend_timeout()')

    def __getitem__(self, item):
        loop = asyncio.get_event_loop()
        try:
            return loop.run_until_complete(self.get(key=item, fail=True))
        except CacheNotFound:
            raise KeyError(f'Key "{item}" not found in cache.')
    
    def __setitem__(self, key, value):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.set(key=key, value=value))
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    def __enter__(self):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.__aenter__())

    def __exit__(self, exc_type, exc_val, exc_tb):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.__aexit__(exc_type, exc_val, exc_tb))




import logging
from datetime import datetime, timedelta
from typing import Any, Optional, Union

from privex.helpers.exceptions import CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT
from privex.helpers.cache.asyncx.base import AsyncCacheAdapter

log = logging.getLogger(__name__)


class AsyncMemoryCache(AsyncCacheAdapter):
    """
    A very basic cache adapter which implements :class:`.AsyncCacheAdapter` - stores the cache in memory using
    the static attribute :py:attr:`.__CACHE`

    As the cache is simply stored in memory, any python object can be cached without needing any form of serialization.

    Fully supports cache expiration.

    **Basic Usage**::

        >>> from time import sleep
        >>> c = AsyncMemoryCache()
        >>> await c.set('test:example', 'hello world', timeout=60)
        >>> await c.get('test:example')
        'hello world'
        >>> sleep(60)
        >>> await c.get('test:example', 'NOT FOUND')
        'NOT FOUND'

    """
    __CACHE = {}

    async def get(self, key: str, default: Any = None, fail: bool = False) -> Any:
        key = str(key)
        c = self.__CACHE
        if key in c:
            log.debug('Cache key "%s" found in __CACHE. Checking expiry...', key)
            vc = c[key]
            if str(vc['timeout']) != 'never' and vc['timeout'] < datetime.utcnow():
                log.debug('Cache key "%s" has expired. Removing from cache.')
                del c[key]
                if fail:
                    raise CacheNotFound(f'Cache key "{key}" was expired.')
                return default
            log.debug('Cache key "%s" is valid and not expired. Returning value "%s"', key, vc)
            return vc['value']
        if fail:
            raise CacheNotFound(f'Cache key "{key}" was not found.')
        log.debug('Cache key "%s" was not found in __CACHE. Returning default value.', key)
        return default

    async def set(self, key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
        key, timeout = str(key), int(timeout)
        c = self.__CACHE
        log.debug('Setting cache key "%s" to value "%s" with timeout %s', key, value, timeout)
        c[key] = dict(value=value, timeout=datetime.utcnow() + timedelta(seconds=timeout))
        return c[key]
    
    async def remove(self, *key: str) -> bool:
        removed = 0
        for k in key:
            k = str(k)
            if k in self.__CACHE:
                del self.__CACHE[k]
                removed += 1
        return removed == len(key)
    
    async def update_timeout(self, key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
        key, timeout = str(key), int(timeout)
        v = await self.get(key=key, fail=True)
        self.__CACHE[key]['timeout'] = datetime.utcnow() + timedelta(seconds=timeout)
        return v



"""
Functions and classes related to working with Python's native asyncio support

To avoid issues with the ``async`` keyword, this file is named ``asyncx`` instead of ``async``

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import asyncio
import inspect
import warnings
from asyncio.subprocess import PIPE, STDOUT
from typing import Tuple, Callable, Any, Union, Coroutine, List, Type

from privex.helpers.common import byteify, shell_quote
from privex.helpers.types import T, STRBYTES, NO_RESULT
import logging

log = logging.getLogger(__name__)


__all__ = [
    'awaitable', 'AWAITABLE_BLACKLIST_MODS', 'AWAITABLE_BLACKLIST', 'AWAITABLE_BLACKLIST_FUNCS', 'run_sync', 'aobject',
    'call_sys_async', 'async_sync', 'awaitable_class', 'AwaitableMixin', 'loop_run', 'is_async_context',
]


def run_sync(func, *args, **kwargs):
    """
    Run an async function synchronously (useful for REPL testing async functions). (TIP: Consider using :func:`.loop_run` instead)

    .. Attention:: For most cases, you should use the function :func:`.loop_run` instead of this. Unlike ``run_sync``, :func:`.loop_run`
                   is able to handle async function references, coroutines, as well as coroutines / async functions which are wrapped
                   in an outer non-async function (e.g. an ``@awaitable`` wrapper).
                   
                   :func:`.loop_run` also supports using a custom event loop, instead of being limited to :func:`asyncio.get_event_loop`

    Usage:

        >>> async def my_async_func(a, b, x=None, y=None):
        ...     return a, b, x, y
        >>>
        >>> run_sync(my_async_func, 1, 2, x=3, y=4)
        (1, 2, 3, 4,)

    :param callable func: An asynchronous function to run
    :param args:          Positional arguments to pass to ``func``
    :param kwargs:        Keyword arguments to pass to ``func``
    """
    coro = asyncio.coroutine(func)
    future = coro(*args, **kwargs)
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(future)
    # return asyncio.run(future)


def loop_run(coro: Union[Coroutine, Type[Coroutine], Callable], *args, _loop=None, **kwargs) -> Any:
    """
    Run the coroutine or async function ``coro`` synchronously, using an AsyncIO event loop.
    
    If the keyword argument ``_loop`` isn't specified, it defaults to the loop returned by :func:`asyncio.get_event_loop`

    If ``coro`` doesn't appear to be a coroutine or async function:
        
    * If ``coro`` is a normal ``callable`` object e.g. a function, then it'll be called.
        
        * If the object returned after calling ``coro(*args, **kwargs)`` **is a co-routine / async func**, then it'll call
          ``loop_run`` again, passing the object returned from calling it, and returning the result from that recursive call.
        
        * If the returned object isn't an async func / co-routine, then the object will be returned as-is.
        
    *  Otherwise, ``coro`` will just be returned back to the caller.
    
    **Example Usage**
    
    First we'll define the async function ``some_func`` to use as an example::
    
        >>> async def some_func(x, y):
        ...     return x + y
    
    Option 1 - Call an async function directly with any args/kwargs required, then pass the coroutine returned::
    
        >>> loop_run(some_func(3, 4))
        7
    
    Option 2 - Pass a reference to the async function, and pass any required args/kwargs straight to :func:`.loop_run` - the
    function will be ran with the args/kwargs you provide, then the coroutine ran in an event loop::
    
        >>> loop_run(some_func, 10, y=20)    # Opt 2. Pass the async function and include any args/kwargs for the call
        30
    
    
    :param coro:     A co-routine, or reference to an async function to be ran synchronously
    :param args:     Any positional arguments to pass to ``coro`` (if it's a function reference and not a coroutine)
    :param _loop:    (kwarg only!) If passed, will run ``coro`` in this event loop, instead of :func:`asyncio.get_event_loop`
    :param kwargs:   Any keyword arguments to pass to ``coro`` (if it's a function reference and not a coroutine)
    
    :type _loop: asyncio.base_events.BaseEventLoop
    
    :return Any coro_result: The returned data from executing the coroutine / async function
    """
    if not asyncio.iscoroutinefunction(coro) and not asyncio.iscoroutine(coro):
        log.debug("Passed 'coro' object isn't a co-routine or async func. Actual type: %s", coro)
        if callable(coro):
            log.debug("Passed 'coro' object is a callable. Calling coro object with args: %s // kwargs: %s", args, kwargs)
            x = coro(*args, **kwargs)
            if asyncio.iscoroutinefunction(x) or asyncio.iscoroutine(x):
                log.debug("Call result from 'coro(*args, **kwargs)' is a coro / async func. Re-running loop_run. ")
                return loop_run(x, *args, _loop=_loop, **kwargs)
            log.debug("Call result from 'coro(*args, **kwargs)' isn't async func / coro. Returning object: %s ", x)
            return x
        log.debug("'coro' object isn't callable. Returning original 'coro' object: %s", coro)
        return coro
    
    loop = _loop
    if loop is None: loop = asyncio.get_event_loop()
    if asyncio.iscoroutinefunction(coro): coro = coro(*args, **kwargs)
    
    return loop.run_until_complete(coro)


def async_sync(f):
    """
    Async Synchronous Decorator, borrowed from https://stackoverflow.com/a/23036785/2648583 - added this PyDoc comment
    and support for returning data from a synchronous function

    Allows a non-async function to run async functions using ``yield from`` - and can also return data

    Useful for unit testing, since unittest.TestCase functions are synchronous.

    Example async function:

        >>> async def my_async_func(a, b, x=None, y=None):
        ...     return a, b, x, y
        ...

    Using the above async function with a non-async function:

        >>> @async_sync
        ... def sync_function():
        ...     result = yield from my_async_func(1, 2, x=3, y=4)
        ...     return result
        ...
        >>> r = sync_function()
        >>> print(r)
        (1, 2, 3, 4,)
        >>> print(r[1])
        2


    """

    def wrapper(*args, **kwargs):
        coro = asyncio.coroutine(f)
        future = coro(*args, **kwargs)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(future)
    return wrapper


async def call_sys_async(proc, *args, write: STRBYTES = None, **kwargs) -> Tuple[bytes, bytes]:
    """
    Async version of :func:`.call_sys` - works exactly the same, other than needing to be ``await``'d.
    Run a process ``proc`` with the arguments ``*args``, optionally piping data (``write``) into the
    process's stdin - then returns the stdout and stderr of the process.
    
    By default, ``stdout`` and ``stdin`` are set to :attr:`asyncio.PIPE` while stderr defaults to
    :attr:`asyncio.STDOUT`. You can override these by passing new values as keyword arguments.
    
    While it's recommended to use the file descriptor types from the :mod:`asyncio` module, they're generally just
    aliases to the types in :mod:`subprocess`, meaning :attr:`subprocess.PIPE` should work the same
    as :attr:`asyncio.PIPE`.
    
    
    **Simple Example**::
        
        >>> from privex.helpers import call_sys_async, stringify
        >>> # All arguments are automatically quoted if required, so spaces are completely fine.
        >>> folders, _ = await call_sys_async('ls', '-la', '/tmp/spaces are fine/hello world')
        >>> print(stringify(folders))
        total 0
        drwxr-xr-x  3 user  wheel  96  6 Dec 17:46 .
        drwxr-xr-x  3 user  wheel  96  6 Dec 17:46 ..
        -rw-r--r--  1 user  wheel   0  6 Dec 17:46 example

    
    **Piping data into a process**::
    
        >>> data = "hello world"
        >>> # The data "hello world" will be piped into wc's stdin, and wc's stdout + stderr will be returned
        >>> out, _ = await call_sys_async('wc', '-c', write=data)
        >>> int(out)
        11
    
    
    
    :param str proc: The process to execute.
    :param str args: Any arguments to pass to the process ``proc`` as positional arguments.
    :param bytes|str write: If this is not ``None``, then this data will be piped into the process's STDIN.
    
    :key stdout: The subprocess file descriptor for stdout, e.g. :attr:`asyncio.PIPE` or :attr:`asyncio.STDOUT`
    :key stderr: The subprocess file descriptor for stderr, e.g. :attr:`asyncio.PIPE` or :attr:`asyncio.STDOUT`
    :key stdin: The subprocess file descriptor for stdin, e.g. :attr:`asyncio.PIPE` or :attr:`asyncio.STDIN`
    :key cwd: Set the current/working directory of the process to this path, instead of the CWD of your calling script.
    
    :return tuple output: A tuple containing the process output of stdout and stderr
    """
    stdout, stderr, stdin = kwargs.pop('stdout', PIPE), kwargs.pop('stderr', STDOUT), kwargs.pop('stdin', PIPE)
    args = [proc] + list(args)
    cmd = shell_quote(*args)
    handle = await asyncio.subprocess.create_subprocess_shell(cmd, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs)
    c = await handle.communicate(input=byteify(write)) if write is not None else await handle.communicate()
    stdout, stderr = c
    return stdout, stderr


AWAITABLE_BLACKLIST_FUNCS: List[str] = []
"""
A list of plain function names - for which :func:`.awaitable` decorated function/methods should always
run synchronously.
"""

AWAITABLE_BLACKLIST_MODS: List[str] = []
"""
A list of fully qualified module names - for which :func:`.awaitable` decorated function/methods should always
run synchronously.
"""

AWAITABLE_BLACKLIST: List[str] = []
"""
A list of fully qualified module paths to functions/methods, if any of these functions/methods call an :func:`.awaitable`
decorated function/method, then the awaitable will be ran synchronously regardless of whether there's an active
AsyncIO context or not.
"""


def is_async_context() -> bool:
    """Returns ``True`` if currently in an async context, otherwise ``False``"""
    try:
        import sniffio
    except ImportError as e:
        raise ImportError(f"is_async_context / @awaitable unavailable - 'sniffio' not installed. Exc: {type(e)} {str(e)} ")
    try:
        # Detect if we're in async context
        sniffio.current_async_library()
        return True
    except sniffio.AsyncLibraryNotFoundError:
        return False


def _awaitable_blacklisted(skip=3) -> bool:
    """
    Returns ``True`` if the caller of the function calling ``_awaitable_blacklisted`` is present in the awaitable
    blacklists such as :attr:`.AWAITABLE_BLACKLIST` - otherwise ``False`` if they're not blacklisted.

    :param int skip: Scan the caller function this far up the stack
                     (2 = callee of _awaitable_blacklisted, 3 = callee of callee #2, 4 = callee of #3, 5 = callee of #4 etc.)

    :return bool is_blacklisted: ``True`` if the calling method/module/function is blacklisted, otherwise ``False``.
    """
    try:
        from privex.helpers.black_magic import calling_module, calling_function, caller_name
        
        # Exact module + function/method path match
        if caller_name(skip=skip) in AWAITABLE_BLACKLIST:
            return True
        # Plain function name match
        if calling_function(skip=skip) in AWAITABLE_BLACKLIST_FUNCS:
            return True
        
        _mod = calling_module(skip=skip)
        # Exact module path match
        if _mod in AWAITABLE_BLACKLIST_MODS:
            return True
        # Sub-modules path match (e.g. if hello.world is blacklisted, then hello.world.example is also blacklisted)
        for _m in AWAITABLE_BLACKLIST_MODS:
            if _mod.startswith(_m + '.'):
                return True
    
    except Exception:
        log.exception("Failed to check blacklist for awaitable function. Falling back to standard async sniffing.")
    
    return False


class AwaitableMixin:
    def __getattribute__(self, item):
        a = object.__getattribute__(self, item)
        
        if not inspect.iscoroutinefunction(a) and not inspect.iscoroutine(a): return a

        def _wrp(*args, **kwargs):
            return a if is_async_context() and not _awaitable_blacklisted() else loop_run(a, *args, **kwargs)

        return _wrp


def awaitable_class(cls: Type[T]) -> Type[T]:
    """
    Wraps a class, allowing all async methods to be used in non-async code as if they were normal synchronous methods.
    
    **Example Usage**
    
    Simply decorate your class with ``@awaitable_class`` (no brackets! takes no arguments), and once you create an instance of your
    class, all of your async methods can be used by synchronous code as-if they were plain functions::
    
        >>> from privex.helpers import awaitable_class
        >>>
        >>> @awaitable_class
        >>> class ExampleAsyncCls:
        >>>     async def example_async(self):
        >>>         return "hello async world"
        >>>
        >>>     def example_sync(self):
        >>>         return "hello non-async world"
        >>>
    
    NOTE - You can also wrap a class without using a decorator - just pass the class as the first argument like so::
        
        >>> class _OtherExample:
        ...     async def hello(self):
        ...         return 'world'
        >>> OtherExample = awaitable_class(_OtherExample)
    
    If we call ``.example_async()`` on the above class from a synchronous REPL, it will return ``'hello async world'`` as if it were a
    normal synchronous method. We can also call the non-async ``.example_sync()`` which works like normal::
    
        >>> k = ExampleAsyncCls()
        >>> k.example_async()
        'hello async world'
        >>> k.example_sync()
        'hello non-async world'
    
    However, inside of an async context (e.g. an async function), ``awaitable_class`` will be returning coroutines, so you should
    ``await`` the methods, as you would expect when dealing with an async function::
        
        >>> async def test_async():
        >>>     exmp = ExampleAsyncCls()
        >>>     return await exmp.example_async()
        >>>
        >>> await test_async()
        'hello async world'

        
    :param type cls: The class to wrap
    :return type wrapped_class: The class after being wrapped
    """
    cls: Type[object]
    
    class _AwaitableClass(cls):
        __AWAITABLE_CLS = True
        """
        This sub-class is modified to appear as if it were the original class being sub-classed, unfortunately this means
        it would be difficult to check whether or not a class has been wrapped with _AwaitableClass.
        
        To allow you to check whether or not a class has been sub-classed by _AwaitableClass, this class private attribute
        is present on the returned class and any instances of it.
        """
        
        def __getattribute__(self, item):
            a: Union[Coroutine, callable, Any] = super().__getattribute__(item)
            cls_name = super().__getattribute__('__class__').__name__
            # full_attr = f"{self.__class__.__name__}.{item}"
            full_attr = f"{cls_name}.{item}"
            if not inspect.iscoroutinefunction(a) and not inspect.iscoroutine(a):
                log.debug("Attribute %s is not a coroutine or coro function. Returning normally.", full_attr)
                return a
            # return awaitable(a) if inspect.iscoroutinefunction(a) else a
            
            def _wrp(*args, **kwargs):
                if is_async_context() and not _awaitable_blacklisted():
                    log.debug("Currently in async context. Returning %s as coroutine", full_attr)
                    return a(*args, **kwargs)
                log.debug("Not in async context or attribute is blacklisted. Returning %s as coroutine", full_attr)
                return loop_run(a, *args, **kwargs)
            return _wrp
    
    _AwaitableClass.__name__ = cls.__name__
    _AwaitableClass.__qualname__ = cls.__qualname__
    _AwaitableClass.__module__ = cls.__module__
    
    return _AwaitableClass


def awaitable(func: Callable) -> Callable:
    """
    Decorator which helps with creation of async wrapper functions.
    
    **Usage**
    
    Define your async function as normal, then create a standard python function using this decorator - the function
    should just call your async function and return it.
    
        >>> async def some_func_async(a: str, b: str):
        ...     c = a + b
        ...     return c
        ...
        >>> @awaitable
        >>> def some_func(a, b) -> Union[str, Coroutine[Any, Any, str]]:
        ...     return some_func_async(a, b)
        ...
    
    Now, inside of async functions, we just ``await`` the wrapper function as if it were the original async function.
    
        >>> async def my_async_func():
        ...     res = await some_func("hello", "world")
        ...
    
    While inside of synchronous functions, we call the wrapper function as if it were a normal synchronous function.
    The decorator will create an asyncio event loop, run the function, then return the result - transparent to the
    calling function.
    
        >>> def my_sync_func():
        ...     res = some_func("hello world")
        ...
    
    **Blacklists**
    
    If you mix a lot of synchronous and asynchronous code, :mod:`sniffio` may return coroutines to synchronous functions
    that were called from asynchronous functions, which can of course cause problems.
    
    To avoid this issue, you can blacklist function names, module names (and their sub-modules), and/or fully qualified
    module paths to functions/methods.
    
    Three blacklists are available in this module, which allow you to specify caller functions/methods, modules, or
    fully qualified module paths to functions/methods for which :func:`.awaitable` wrapped functions/methods
    should **always** execute in an event loop and return synchronously.
    
    Example::
        
        >>> from privex.helpers import asyncx
        >>> # All code within the module 'some.module' and it's sub-modules will always have awaitable's run their wrapped
        >>> # functions synchronously.
        >>> asyncx.AWAITABLE_BLACKLIST_MODS += ['some.module']
        >>> # Whenever a function with the name 'example_func' (in any module) calls an awaitable, it will always run synchronously
        >>> asyncx.AWAITABLE_BLACKLIST_FUNCS += ['example_func']
        >>> # Whenever the specific class method 'other.module.SomeClass.some_sync' calls an awaitable, it will always run synchronously.
        >>> asyncx.AWAITABLE_BLACKLIST += ['other.module.SomeClass.some_sync']
    
    
    Original source: https://github.com/encode/httpx/issues/572#issuecomment-562179966
    
    """
    def wrapper(*args: Any, **kwargs: Any) -> Union[Any, Coroutine[Any, Any, Any]]:
        coroutine = func(*args, **kwargs)

        # The wrapped function isn't a coroutine function, nor a coroutine. This may be caused by an adapter
        # wrapper class which deals with both synchronous and asynchronous adapters.
        # Since it doesn't appear to be a coroutine, just return the result.
        if not asyncio.iscoroutinefunction(coroutine) and not asyncio.iscoroutine(coroutine):
            return coroutine

        # Always run the coroutine in an event loop if the caller function is blacklisted in the AWAITABLE_BLACKLIST* lists
        if _awaitable_blacklisted():
            return asyncio.get_event_loop().run_until_complete(coroutine)
            
        if is_async_context():           # We're in async context, return the coroutine for await usage
            return coroutine
        loop = asyncio.get_event_loop()  # Not in async context, run coroutine in event loop.
        return loop.run_until_complete(coroutine)
    
    return wrapper


# noinspection All
class aobject(object):
    """
    Inheriting this class allows you to define an async __init__.

    To use async constructors, you must construct your class using ``await MyClass(params)``
    
    **Example**::
        
        >>> class SomeClass(aobject):
        ...     async def __init__(self, some_param='x'):
        ...         self.some_param = some_param
        ...         self.example = await self.test_async()
        ...
        ...     async def test_async(self):
        ...         return "hello world"
        ...
        >>> async def main():
        ...     some_class = await SomeClass('testing')
        ...     print(some_class.example)
        ...
    
    **Note:** Some IDEs like PyCharm may complain about having async ``__new__`` and ``__init__``, but it **does**
    work with Python 3.6+.
    
    You may be able to work-around the syntax error in your sub-class by defining your ``__init__`` method under a
    different name, and then assigning ``__init__ = _your_real_init`` much like this class does.
    
    Original source: https://stackoverflow.com/a/45364670
    """

    # noinspection PyUnresolvedReferences
    async def __aobject_new(cls, *a, **kw):
        instance = super().__new__(cls)
        # noinspection PyArgumentList
        await instance.__init__(*a, **kw)
        return instance

    async def __aobject_init(self):
        pass

    __new__ = __aobject_new
    __init__ = __aobject_init





from os import getcwd
from os.path import isabs, abspath, join
from typing import Tuple, Optional, Coroutine, Union, Any

from privex.helpers.common import stringify, empty, empty_if
from privex.helpers import STRBYTES
from privex.helpers.asyncx import call_sys_async, awaitable, awaitable_class
from privex.helpers.exceptions import SysCallError


async def _async_sys(proc, *args, write: STRBYTES = None, **kwargs) -> Tuple[str, str]:
    stderr_raise = kwargs.pop('stderr_raise', False)
    strip = kwargs.pop('strip', True)
    out, err = await call_sys_async(proc, *args, write=write, **kwargs)
    out, err = stringify(out, if_none=''), stringify(err, if_none='')
    if strip: out, err = out.strip(), err.strip()
    
    if len(err) > 1 and stderr_raise:
        raise SysCallError(f'Git returned an error: "{err}"')
    
    return out, err


def _repo(repo):
    repo = getcwd() if empty(repo) else repo
    return join(getcwd(), repo) if not isabs(repo) else repo


@awaitable_class
class AsyncGit:
    """
    This class uses the :func:`.awaitable_class` decorator to make the class work both synchronously and asynchronously, without needing
    to duplicate code.
    
    The :func:`.awaitable_class` decorator detects whether the methods are being called from a synchronous, or an asynchronous
    function/method.
    
    * If they're being called from an async function, then a coroutine will be returned, and must be ``await``'ed.
    
    * If they're being called from a synchronous function, then it will spin up an event loop, run the method in
      the event loop, then return the result transparently.
    
    
    **Synchronous usage**::
        
        >>> g = Git()   # Git is an alias for AsyncGit, there is no difference between them
        >>>
        >>> def some_func():
        ...     current_commit = g.get_current_commit()
        ...     print(current_commit)
        'ac52b28f551825160785f9ea7e96f86ccc869cc1'
    
    **Asynchronous usage**::
        
        >>> g = Git()    # Git is an alias for AsyncGit, there is no difference between them
        >>>
        >>> async def some_func():
        ...     current_commit = await g.get_current_commit()
        ...     print(current_commit)
        'ac52b28f551825160785f9ea7e96f86ccc869cc1'

    """
    repo: Optional[str]
    default_version: str
    
    def __init__(self, repo: str = None, default_version: str = 'HEAD', **kwargs):
        self.repo = None
        self.repo = _repo(repo)
        self.default_version = default_version
    
    def _repo(self, repo=None) -> str:
        return self.repo if empty(repo) and not empty(self.repo) else _repo(repo)

    async def _git(self, *args, repo: str = None, strip=True) -> str:
        repo = self._repo(repo)
        out, _ = await _async_sys("git", *args, cwd=repo, strip=strip)
        return out
    
    async def init(self, *args, repo: str = None) -> str:
        return await self._git("init", *args, repo=repo)

    async def add(self, *args, repo: str = None) -> str:
        return await self._git("add", *args, repo=repo)

    async def branch(self, *args, repo: str = None) -> str:
        return await self._git("branch", *args, repo=repo)

    async def commit(self, message: str, *args, repo: str = None) -> str:
        return await self._git("commit", "-m", stringify(message), *args, repo=repo)

    async def checkout(self, branch: str, *args, repo: str = None, new: bool = False) -> str:
        args = ['-b', branch] + list(args) if new else [branch] + list(args)
        return await self._git("checkout", *args, repo=repo)
    
    async def status(self, *args, repo: str = None, concise=True) -> str:
        if concise: args = ['-s'] + list(args)
        return await self._git("status", *args, repo=repo, strip=False)

    async def tag(self, *args, repo: str = None) -> str:
        return await self._git("tag", *args, repo=repo)
    
    async def get_current_commit(self, version: str = None, repo: str = None) -> str:
        """
        Get current commit hash. Optionally specify ``version`` to get the current commit hash for a branch or tag.
        
        **Examples**::
        
            >>> g = Git()
            >>> await g.get_current_commit()
            'ac52b28f551825160785f9ea7e96f86ccc869cc1'
            >>> await g.get_current_commit('2.0.0')    # Get the commit hash for the tag 2.0.0
            '598584a447ba63212ac3fe798c01941badf1c194'
        
        
        :param str version:  Optionally specify a branch / tag to get the current commit hash for.
        :param str repo:     Optionally specify a specific local repository path to run ``git`` within.
        :return str commit_hash: The current Git commit hash
        """
        return await self._git("rev-parse", empty_if(version, self.default_version), repo=repo)
    
    async def get_current_branch(self, repo: str = None) -> str:
        """
        Get current active branch/tag.

        **Examples**::

            >>> g = Git()
            >>> await g.get_current_branch()
            'master'
            >>> await g.checkout('testing', new=True)    # Create and checkout the branch 'testing'
            >>> await g.get_current_branch()
            'testing'
        

        :param str repo:     Optionally specify a specific local repository path to run ``git`` within.
        :return str current_branch: The name of the current checked out branch or tag
        """
        return await self._git("rev-parse", "--abbrev-ref", 'HEAD', repo=repo)

    async def get_current_tag(self, version: str = None, repo: str = None) -> str:
        """
        Get the latest tag on this branch - useful for detecting current version of your python application.

        **Examples**::

            >>> g = Git()
            >>> await g.get_current_tag()          # Get the latest tag on the active branch
            '2.5.0'
            >>> await g.get_current_tag('develop') # Get the latest tag on the branch 'develop'
            '2.5.3'

        :param str version:  Optionally specify a branch / tag to get the latest tag for.
        :param str repo:     Optionally specify a specific local repository path to run ``git`` within.
        :return str current_tag: The name of the latest tag on this branch.
        """
        return await self._git("describe", "--abbrev=0", "--tags", empty_if(version, self.default_version), repo=repo)
    
    async def log(self, *args, repo: str = None, concise=True):
        if concise: args = ['--oneline'] + list(args)
        return await self._git("--no-pager", "log", *args, repo=repo)


Git = AsyncGit
_cwd_git = AsyncGit()

get_current_commit = _cwd_git.get_current_commit
get_current_branch = _cwd_git.get_current_branch
get_current_tag = _cwd_git.get_current_tag

__all__ = [
    'AsyncGit', 'Git', 'get_current_commit', 'get_current_branch', 'get_current_tag', '_repo',
]



"""
This module handles connection objects for databases, APIs etc. by exposing functions which initialise and
store class instances for re-use.

It's primarily intended to be used to enable database, caching and third-party API connectivity for the helpers in this
package, however, you're free to use the functions / classes / attributes exposed in this module for your own apps.

Classes are generally initialised using the settings from :py:mod:`.settings` - see the docs for that module to learn
how to override the settings if the defaults don't work for you.


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import logging
import threading
from typing import Any, Union, Optional

from privex.helpers import settings
from privex.helpers.types import T
from privex.helpers.common import empty_if

log = logging.getLogger(__name__)

__all__ = [
    'HAS_REDIS', 'HAS_ASYNC_REDIS', 'HAS_DNSPYTHON', 'HAS_CRYPTO', 'HAS_SETUPPY_BUMP',
    'HAS_SETUPPY_COMMANDS', 'HAS_SETUPPY_COMMON'
]

HAS_REDIS = False
"""If the ``redis`` module was imported successfully, this will change to True."""

HAS_ASYNC_REDIS = False
"""If the ``aioredis`` module was imported successfully, this will change to True."""

HAS_DNSPYTHON = False
"""If the ``dns.resolver`` module was imported successfully, this will change to True."""

HAS_CRYPTO = False
"""If :py:mod:`privex.helpers.crypto` was imported successfully, this will change to True"""

HAS_SETUPPY_COMMON = False
"""If :py:mod:`privex.helpers.setuppy.common` was imported successfully, this will change to True"""

HAS_SETUPPY_BUMP = False
"""If :py:mod:`privex.helpers.setuppy.bump` was imported successfully, this will change to True"""

HAS_SETUPPY_COMMANDS = False
"""If :py:mod:`privex.helpers.setuppy.commands` was imported successfully, this will change to True"""

__STORE = dict(threads={})
"""This ``dict`` is used to store initialised classes for connections to databases, APIs etc."""


def _get_threadstore(name=None, fallback=None, thread_id=None) -> Any:
    thread_id = empty_if(thread_id, threading.get_ident())
    if thread_id not in __STORE['threads']:
        __STORE['threads'][thread_id] = {}
    thread_store: dict = __STORE['threads'][thread_id]
    if name is None:
        return thread_store
    
    return thread_store.get(name, fallback)


def _set_threadstore(name, obj: T, thread_id=None) -> T:
    thread_store = _get_threadstore(thread_id=thread_id)
    thread_store[name] = obj
    return obj


def clean_threadstore(thread_id=None, name=None):
    """
    Remove the per-thread instance storage in :attr:`.__STORE`, usually called when a thread is exiting.

    Example::

        >>> def some_thread():
        ...     r = get_redis()
        ...     print('doing something')
        ...     print('cleaning up...')
        ...     clean_threadstore()       # With no arguments, it cleans the thread store for the thread that called it.
        >>> t = threading.Thread(target=some_thread)
        >>> t.start()
        >>> t.join()

    Usage outside of a thread::

        >>> t = threading.Thread(target=some_thread)
        >>> t.start()
        >>> thread_id = t.ident                      # Get the thread ID for the started thread
        >>> t.join()                                 # Wait for the thread to finish
        >>> if thread_id is not None:                # Make sure the thread ID isn't None
        ...     clean_threadstore(thread_id)         # Cleanup any leftover instances, if there are any.
        ...

    Removing an individual item from thread store::
        
        >>> def some_thread():
        ...     r = get_redis()
        ...     print('doing something')
        ...     print('cleaning up...')
        ...     clean_threadstore(name='redis')   # Delete only the key 'redis' from the thread store
    

    :param thread_id: The ID of the thread (usually from :func:`threading.get_ident`) to clean the storage for.
                      If left as None, will use the ID returned by :func:`threading.get_ident`.
    
    :param name:      If specified, then only the key ``name`` will be deleted from the thread store, instead of the entire thread store.
    """
    thread_id = empty_if(thread_id, threading.get_ident())
    if thread_id in __STORE['threads']:
        ts = __STORE['threads'][thread_id]
        if name is None:
            del __STORE['threads'][thread_id]
            return True
        elif name in ts:
            del ts[name]
            return True
    return False


try:
    import redis


    def connect_redis(*args, **rd_config):
        from privex.helpers.common import extract_settings
        # return redis.Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT, db=settings.REDIS_DB)
        return redis.Redis(*args, **extract_settings('REDIS_', settings, merge_conf=dict(rd_config)))
    
    def get_redis(new_connection=False, thread_id=None, **rd_config) -> redis.Redis:
        """Get a Redis connection object. Create one if it doesn't exist."""
        from privex.helpers.common import empty_if, extract_settings

        if new_connection:
            return connect_redis(**rd_config)
        # if thread_id not in __STORE['threads']: __STORE['threads'][thread_id] = {}

        # thread_store = __STORE['threads'][thread_id]
        rd = _get_threadstore('redis', thread_id=thread_id)
        
        if rd is None:
            rd = connect_redis(**rd_config)
            _set_threadstore('redis', rd, thread_id=thread_id)
        return rd
        # if 'redis' in thread_store and not new_connection: return thread_store['redis']
        # if rd is not None and not new_connection:
        #
        # # if 'redis' not in thread_store:
        #     if new_connection:
        #         return rd
        #     thread_store['redis'] = rd
        # return thread_store['redis']
    
    def close_redis(thread_id=None) -> bool:
        """
        Close the global Redis connection and delete the instance.
        
        :param thread_id: Close and delete the Redis instance for this thread ID, instead of the detected current thread
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Redis instance.
        """
        
        rd: Union[str, redis.Redis] = _get_threadstore('redis', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            rd.close()
            clean_threadstore(thread_id=thread_id, name='redis')
            return True
        return False

    def reset_redis(thread_id=None) -> redis.Redis:
        """Close the global Redis connection, delete the instance, then re-instantiate it"""
        close_redis(thread_id=thread_id)
        return _set_threadstore('redis', get_redis(True, thread_id), thread_id=thread_id)
    
    def configure_redis(host=settings.REDIS_HOST, port: int = settings.REDIS_PORT, db: int = settings.REDIS_DB, **kwargs):
        """Update global Redis settings and re-instantiate the global Redis instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.REDIS_DB = db
        settings.REDIS_PORT = port
        settings.REDIS_HOST = host
        return reset_redis(thread_id=thread_id)


    __all__ += ['get_redis', 'reset_redis', 'configure_redis', 'close_redis', 'connect_redis']
    HAS_REDIS = True

except ImportError:
    log.debug('privex.helpers __init__ failed to import "redis", Redis dependent helpers will be disabled.')
except Exception as e:
    log.debug('privex.helpers __init__ failed to import "redis", (unknown exception), '
              'Redis dependent helpers will be disabled. Exception: %s %s', type(e), str(e))

try:
    import aioredis
    HAS_ASYNC_REDIS = True


    async def connect_redis_async(**rd_config) -> aioredis.Redis:
        from privex.helpers.common import extract_settings
        cf = extract_settings('REDIS_', settings, merge_conf=dict(rd_config))
        _addr = f"redis://{cf.pop('host', 'localhost')}:{cf.pop('port', 6379)}"
        return await aioredis.create_redis_pool(
            address=_addr, **cf
        )


    async def get_redis_async(new_connection=False, thread_id=None, **rd_config) -> aioredis.connection:
        """
        Get an Async Redis connection object. Create one if it doesn't exist.
        
        Usage::
            
            >>> redis_conn = await get_redis_async()
            >>> redis = await redis_conn
            >>> await redis.set('some_key', 'example')
            >>> await redis.get('some_key')
            'example'
            
        
        """
        from privex.helpers.common import empty_if, extract_settings
    
        if new_connection: return await connect_redis_async(**rd_config)
        rd: Optional[aioredis.Redis] = _get_threadstore('aioredis', thread_id=thread_id)
    
        if rd is None:
            rd: aioredis.Redis = _set_threadstore('aioredis', await connect_redis_async(**rd_config), thread_id=thread_id)
        return rd

    async def close_redis_async(thread_id=None) -> bool:
        """
        Close the global Async Redis connection and delete the instance.

        :param thread_id: Close and delete the Redis instance for this thread ID, instead of the detected current thread
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Redis instance.
        """
    
        rd: Union[str, aioredis.connection] = _get_threadstore('aioredis', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            rd.close()
            await rd.wait_close()
            clean_threadstore(thread_id=thread_id, name='aioredis')
            return True
        return False


    async def reset_redis_async(thread_id=None) -> aioredis.connection:
        """Close the global Async Redis connection, delete the instance, then re-instantiate it"""
        await close_redis_async(thread_id=thread_id)
        return await _set_threadstore('aioredis', get_redis_async(True, thread_id), thread_id=thread_id)


    async def configure_redis_async(host=settings.REDIS_HOST, port: int = settings.REDIS_PORT, db: int = settings.REDIS_DB, **kwargs):
        """Update global Redis settings and re-instantiate the global Async Redis instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.REDIS_DB = db
        settings.REDIS_PORT = port
        settings.REDIS_HOST = host
        return await reset_redis_async(thread_id=thread_id)


    __all__ += ['get_redis_async', 'reset_redis_async', 'configure_redis_async', 'close_redis_async', 'connect_redis_async']


except ImportError:
    log.debug('privex.helpers __init__ failed to import "aioredis", Async Redis dependent helpers will be disabled.')
except Exception as e:
    log.debug('privex.helpers __init__ failed to import "aioredis", (unknown exception), '
              'Async Redis dependent helpers will be disabled. Exception: %s %s', type(e), str(e))

try:
    import aiomcache
    HAS_ASYNC_MEMCACHED = True

    async def connect_memcached_async(**rd_config) -> aiomcache.Client:
        from privex.helpers.common import extract_settings
        cf = extract_settings('MEMCACHED_', settings, merge_conf=dict(rd_config))
        # if 'loop' not in cf:
        #     cf['loop'] = asyncio.get_event_loop()
        return aiomcache.Client(**cf)


    async def get_memcached_async(new_connection=False, thread_id=None, **rd_config) -> aiomcache.Client:
        """Get an Async Memcached connection object. Create one if it doesn't exist."""
        from privex.helpers.common import empty_if, extract_settings
    
        if new_connection: return await connect_memcached_async(**rd_config)
        rd: Optional[aiomcache.Client] = _get_threadstore('aiomemcached', thread_id=thread_id)
    
        if rd is None:
            rd: aiomcache.Client = _set_threadstore('aiomemcached', await connect_memcached_async(**rd_config), thread_id=thread_id)
        return rd


    def close_memcached_async(thread_id=None) -> bool:
        """
        Close the global Async Memcached connection and delete the instance.

        :param thread_id: Close and delete the Memcached instance for this thread ID, instead of the detected current thread
        :return bool deleted: ``True`` if an instance was found and deleted. ``False`` if there was no existing Memcached instance.
        """
    
        rd: Union[str, aiomcache.Client] = _get_threadstore('aiomemcached', 'NOT_FOUND', thread_id=thread_id)
        if rd != 'NOT_FOUND':
            rd.close()
            clean_threadstore(thread_id=thread_id, name='aiomemcached')
            return True
        return False


    async def reset_memcached_async(thread_id=None) -> aiomcache.Client:
        """Close the global Async Memcached connection, delete the instance, then re-instantiate it"""
        close_memcached_async(thread_id=thread_id)
        return await _set_threadstore('aiomemcached', get_memcached_async(True, thread_id), thread_id=thread_id)


    async def configure_memcached_async(host=settings.MEMCACHED_HOST, port: int = settings.MEMCACHED_PORT, **kwargs):
        """Update global Memcached settings and re-instantiate the global Async Redis instance with the new settings."""
        thread_id = kwargs.get('thread_id', None)
        settings.MEMCACHED_PORT = port
        settings.MEMCACHED_HOST = host
        return await reset_memcached_async(thread_id=thread_id)


    __all__ += ['get_memcached_async', 'reset_memcached_async', 'configure_memcached_async', 'close_memcached_async', 'connect_memcached_async']

except ImportError:
    log.debug('%s failed to import "aiomcache", Async Memcached dependent helpers will be disabled.', __name__)
except Exception as e:
    log.debug('%s failed to import "aiomcache", (unknown exception), Async Memcached dependent helpers '
              'will be disabled. Exception: %s %s', __name__, type(e), str(e))

